/*
   File Name:

      RayTracer.cpp

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This file has the source to the ray tracing functions used to create the scene.
*/


#include"RayTracer.h"


CVector4* GenerateRayDirections(int width, int height, int depth)
{
   // Create variables to hold half the screen width and height.
   float halfWidth = (float)(width >> 1);
   float halfHeight = (float)(height >> 1);
   
   // Allocate memory for the list of directions.
   CVector4 *direction = new CVector4[width * height];


   // Loop through every pixel of the screen and create the directions for the rays.
   // The ray will go from its source to each pixel on the screen so we need a direction
   // for every pixel.
   for(int y = 0; y < height; y++)
      {
         for(int x = 0; x < width; x++)
            {
               // Get address of the direction array element we are currently
               // working with to make setting the direction data using currentDir.
               CVector4 &currentDir = direction[x + y * width];

               // Set the current direction as x = x iter - half the width, y = y iter
               // - half the height, and what ever depth value we send in.  The last one
               // is the back plane value (distance from camera to has far as one can see).
               // The direction are created starting from the lower left.
               currentDir = CVector4(x - halfWidth, y - halfHeight, (float)depth);
               
               // We must normalize the direction.
               currentDir.Normal();
            }
      }

   // return our newly created ray directions array.
   return direction;
}


void TraceScene(Shape *shapeList, int totalShapes,  CLight &light,
                unsigned char *buffer, CVector4 *directionList, int w, int h)
{
   // Final pixel color, primary ray.
   float r = 0, g = 0, b = 0;
   Ray primaryRay;

   // Pixel color component counter used in the nested for loop.
   int k = 0;
   
   // Starting point for the ray.
   primaryRay.origin.x = 0;
   primaryRay.origin.y = 100;
   primaryRay.origin.z = -600;

   // Loop through the pixels of the backbuffer and set each one based on the
   // color of the shape as the ray hits it or leave it if there is no collision.
   for(int y = 0; y < h; y++)
      {
         for(int x = 0; x < w; x++, k+=3)
            {
               // Grab the direction for this current pixel.
               primaryRay.direction = directionList[x + y * w];

               // This will hold the result of the test of each shape against the ray.
               TraceResult intersectResult;
               
               // The first is set to a really high default number and the
               // second is set to -1 which means no spheres has been intersected.
               float closestIntersectedDist = 1000000;
               int closestIntersectedShape = -1;

               // Loop through all shapes and determine which hits this ray.
               // We will take the closest shape.
               for(int i = 0; i < totalShapes; i++)
               {
                  // Test if this ray hits this shape while going towards this pixel.
                  if(shapeList[i].type == SPHERE && shapeList[i].sphere != 0)
                     intersectResult = shapeList[i].sphere->Intersect(primaryRay);
                  else if(shapeList[i].type == TRIANGLE && shapeList[i].triangle != 0)
                     intersectResult = shapeList[i].triangle->Intersect(primaryRay);
                  else if(shapeList[i].type == PLANE && shapeList[i].plane != 0)
                     intersectResult = shapeList[i].plane->Intersect(primaryRay);

                  // If the ray hits this object...
                  if(intersectResult.hit == true)
                     {
                        // If this distance of this object is closer to the
                        // camera than the last object then we record it.
                        if(intersectResult.distance < closestIntersectedDist)
                           {
                              closestIntersectedDist = intersectResult.distance;
                              closestIntersectedShape = i;
                           }
                     }
               }

            // If no shapes hit this ray then we leave this pixel as
            // the default color.  Else set to shape color.
            if(closestIntersectedShape != -1)
               {
                  // Calculated the shape color with the light shining on it
                  // and store that color in r, g, b.
                  LightShadeShape(shapeList, closestIntersectedShape, totalShapes,
                                  primaryRay, closestIntersectedDist, light, r, g, b);

                  // Save the color in our back buffer array.
                  buffer[k + 0] = (int)r;
                  buffer[k + 1] = (int)g;
                  buffer[k + 2] = (int)b;
               }
         }
      }
}


void LightShadeShape(Shape *shape, int shapeIndex, int totalShapes, Ray &ray,
                     float dist, CLight &light, float &r, float &g, float &b)
{
   CVector4 RaySphereIntersect;

   // Get the interection between the shape and the ray.  We take the origin then
   // add the distance multiplied by the direction the ray is pointing.
   RaySphereIntersect.x = ray.origin.x + dist * ray.direction.x;
   RaySphereIntersect.y = ray.origin.y + dist * ray.direction.y;
   RaySphereIntersect.z = ray.origin.z + dist * ray.direction.z;

   CVector4 normal;

   if(shape[shapeIndex].type == SPHERE && shape[shapeIndex].sphere != 0)
      normal = shape[shapeIndex].sphere->GetNormal(RaySphereIntersect);
   else if(shape[shapeIndex].type == TRIANGLE && shape[shapeIndex].triangle != 0)
      normal = shape[shapeIndex].triangle->GetNormal();
   else if(shape[shapeIndex].type == PLANE && shape[shapeIndex].plane != 0)
      normal = shape[shapeIndex].plane->GetNormal();

   // We need the light direction for the lighting calculations.  lightDist
   // is used to store the distance the light is from the intersected point.
   // This will be used in the shadow calculations.
   CVector4 lightDir;
   float lightDist = 0;

   // Get the light direction from the light pos and the point of intersection.
   lightDir = light.position - RaySphereIntersect;

   // Normalize the light direction.
   lightDist = lightDir.GetLength();
   lightDir.Normal();
   
   // Now we take our data and test if this intersected point is in shadow.
   // If so then the final color is black, else we continue with the lighting.
   Ray shadowRay;
   TraceResult shadowIntersect;
   shadowRay.origin = RaySphereIntersect;
   shadowRay.direction = lightDir;
   
   // Test to see if the point is in a shadow.
   for(int i = 0; i < totalShapes; i++)
      {
         if(shape[i].type == SPHERE && shape[i].sphere != 0)
            shadowIntersect = shape[i].sphere->Intersect(shadowRay);
         else if(shape[i].type == TRIANGLE && shape[i].triangle != 0)
            shadowIntersect = shape[i].triangle->Intersect(shadowRay);
         else if(shape[i].type == PLANE && shape[i].plane != 0)
            shadowIntersect = shape[i].plane->Intersect(shadowRay);

         // If the ray hits this object...
         if(shadowIntersect.hit == true)
            {
               if(shadowIntersect.distance < lightDist)
                  {
                     r = g = b = 0;
                     return;
                  }
            }
      }

   // Calculate the light coefficient.
   float lightCoef;
   lightCoef = lightDir.DotProduct3(normal);

   // If the light coefficient is less than 0 make it = to 0.   
   if(lightCoef < 0) lightCoef = 0;

   // Calculate the final color by the shape color times the light coefficient.
   // First we will save the shape and light colors.  We are dividing by 255
   // to convert the color between 0 to 1.  We will expand this back into 0 to 255
   // after the lighting calculations have taken place.
   float sr = 0, sg = 0, sb = 0;

   if(shape[shapeIndex].type == SPHERE && shape[shapeIndex].sphere != 0)
      {
         sr = shape[shapeIndex].sphere->r / 255;
         sg = shape[shapeIndex].sphere->g / 255;
         sb = shape[shapeIndex].sphere->b / 255;
      }
   else if(shape[shapeIndex].type == TRIANGLE && shape[shapeIndex].triangle != 0)
      {
         sr = shape[shapeIndex].triangle->r / 255;
         sg = shape[shapeIndex].triangle->g / 255;
         sb = shape[shapeIndex].triangle->b / 255;
      }
   else if(shape[shapeIndex].type == PLANE && shape[shapeIndex].plane != 0)
      {
         sr = shape[shapeIndex].plane->r / 255;
         sg = shape[shapeIndex].plane->g / 255;
         sb = shape[shapeIndex].plane->b / 255;
      }

   float lr = light.r / 255;
   float lg = light.g / 255;
   float lb = light.b / 255;
   
   // finalLight* will hold the light contribution.
   float finalLightR = lr * lightCoef;
   float finalLightG = lg * lightCoef;
   float finalLightB = lb * lightCoef;

   // Expand back to 0 to 255 after combining the shape color with the light color.
   r = (sr * finalLightR) * 255;
   g = (sg * finalLightG) * 255;
   b = (sb * finalLightR) * 255;

   // Clamp to keep color values between 0 and 255.
   if(r > 255) r = 255;
   if(g > 255) g = 255;
   if(b > 255) b = 255;
}


// Copyright February 2005
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com