/*
   Ray Tracing Shadows Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to build off of the previous ray tracing demos
   and add shadows to the scene.  Since we already have a way to compute lighting
   in the scene, the next step is to do shadows.  With shadows in ray tracing they
   are pretty much free.  What this means is that you already have all the information
   neccessary to test if a pixel is in shadow or not.  All you have to do is do the
   simple test, if it is in shadow then the color is black (normally).  Else you
   continue with the normal lighting calculations.  All the new code is in the
   updated function LightShadeShape().
*/


// Demo: Ray Tracing Shadows.


// All the necessary includes...
#include<iostream>
#include"Shape.h"
#include"RayTracer.h"
#include"Light.h"
#include"TgaImage.h"

using namespace std;

// Width/Height resolution and total number of objects.
#define WIDTH_RES 640
#define HEIGHT_RES 480
#define TOTAL_SHAPES 4


// Function prototype...
bool InitializeRT();
void ClearBuffer(unsigned char *buffer, int size, int r, int g, int b);
void Shutdown();


// Primary render array and the look up array of directions.
unsigned char *primaryBuffer = 0;
CVector4 *directionList = 0;

// List of objects and light source.
Shape shapes[TOTAL_SHAPES];
CLight light;


int main()
{
   cout << "Ultimate Game Programming Ray Tracer..." << endl;
   cout << "Created by Allen Sherrod..." << endl << endl;

   // Initialize all arrays and data nescessary.  If it fails then we can't
   // perform our ray tracing.
   if(!InitializeRT())
      {
         cout << "Initialization failed!  Exiting." << endl;
         return 0;
      }

   cout << "Clearing primary buffer..." << endl;

   // Clear the rendered scene to a default color (I choose black).
   ClearBuffer(primaryBuffer, WIDTH_RES * HEIGHT_RES * 3, 0, 0, 0);
   
   cout << "Primary buffer cleared and ready to be drawn to..." << endl << endl;
   cout << "Starting trace." << endl;

   // Trace through the scene and create the rendered scene.
	TraceScene(shapes, TOTAL_SHAPES, light, primaryBuffer, directionList,
              WIDTH_RES, HEIGHT_RES);

   cout << "Trace over." << endl << endl;
   
   cout << "Preparing to save scene as RayTracedScene.tga..." << endl;
   
   // Take our ray traced scene and save it as a .tga image file.
   WriteTGA("RayTracedScene.tga", WIDTH_RES, HEIGHT_RES, primaryBuffer);

   cout << "Traced scene saved...shut down starting..." << endl << endl;

   // Release all resources and data we used.
   Shutdown();

   cout << "Exiting..." << endl;

   return 1;
}


bool InitializeRT()
{
   // Release all previous data.
   Shutdown();

   // Allocate rendering buffer.
	primaryBuffer = new unsigned char[WIDTH_RES * HEIGHT_RES * 3];

   // Error checking...
   if(primaryBuffer == NULL)
      return false;

	// Allocate the ray directions.
	directionList = GenerateRayDirections(WIDTH_RES, HEIGHT_RES, 255);

   // Error checking...
   if(directionList == NULL)
      return false;

   // Setup the light.
   light.position = CVector4(200, 100, -600);
   light.r = 255; light.g = 255; light.b = 255;

   // Initialize the shapes we are using.
   shapes[0].type = SPHERE;
   shapes[0].sphere = new Sphere(-150, -50, 0, 100, 128, 75, 255);

   shapes[1].type = SPHERE;
   shapes[1].sphere = new Sphere(150, -50, 0, 100, 255, 128, 128);
   
   shapes[2].type = PLANE;
   shapes[2].plane = new Plane(CVector4(0, -150, 0),
                               CVector4(0, 1, 0), 255, 255, 255);

   shapes[3].type = PLANE;
   shapes[3].plane = new Plane(CVector4(0, 0, 500),
                               CVector4(0, 0, -1), 255, 255, 255);

   return true;
}


void ClearBuffer(unsigned char *buffer, int size, int r, int g, int b)
{
   // Loop through the entire array and set the buffer to the specified color.
   for(int p = 0; p < size; p += 3)
      {
         buffer[p + 0] = r; buffer[p + 1] = g; buffer[p + 2] = b;
      }
}


void Shutdown()
{
   // Release all resources we used.
   if(directionList != NULL) delete[] directionList;
   if(primaryBuffer != NULL) delete[] primaryBuffer;


   for(int i = 0; i < TOTAL_SHAPES; i++)
      {
         if(shapes[i].type == SPHERE && shapes[i].sphere != NULL)
            delete shapes[i].sphere;
         else if(shapes[i].type == TRIANGLE && shapes[i].triangle != NULL)
            delete shapes[i].triangle;
         else if(shapes[i].type == PLANE && shapes[i].plane != NULL)
            delete shapes[i].plane;
      }
}


// Recap:

// Shadows add a great deal of realism to any scene.  Even though these shadows
// are hard (soft shadows are more realistic) the scene still looks great.  What
// is good about ray traced shadows is that they come practically for free.
// Create the shadow ray is nothing more than using data we already computed.
// Once we have the shadow ray we can just test if it intersects any objects
// in the scene using the intersect functions of each shape.  If there is
// intersection and the object is in front of the light, we are in a shadow.


// Copyright February 2005
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com